clc;
clear all;
%% Model
M = 100;   % total number of timesteps
N = 5;     % horizon length

nx = 2;
ny = 2;

A = [0.9 0; 0 0.8];
B = [1 0; 0 1];

% model equations and corresponding Jacobians
f = @(x) A*x;
h = @(x) B*x;

J_f = @(x) A;
J_h = @(x) B;

Q = 0.001*eye(nx);   % state noise covariance
R = 0.2*eye(ny);     % measurement noise covariance

x0 = [2; 1];         % prior mean
Q0 = 0.1*eye(nx);    % prior covariance

%% Load data

load('data_lin.mat');

% plot true state and measurements
figure(1);
subplot(2, 1, 1); grid on; hold on;
plot(1:M, X(1, :), '.-');
plot(1:M, X(2, :), '.-');
legend('x1', 'x2');
xlabel('timestep k');

subplot(2, 1, 2); grid on; hold on;
plot(1:M, Y(1, :), '.-');
plot(1:M, Y(2, :), '.-');
legend('y1', 'y2');
xlabel('timestep k');

%% Kalman Filter

x_estimate_kf = zeros(nx, M);
x_predict = zeros(nx, M);         % used as arrival cost mean
P_predict = zeros(nx, nx, M);     % used as arrival cost covariance

x_ = x0;
P_ = Q0;

for k=1:M
    x_predict(:, k) = x_;
    P_predict(:, :, k) = P_;
    
    [x, P] = update_ekf(Y(:, k), x_, P_, h, J_h, R);
    [x_, P_] = predict_ekf(x, P, f, J_f, Q);
    
    x_estimate_kf(:, k) = x;
end


%% Moving Horizon Estimation

x_estimate_mhe = zeros(nx, M-N);

% get initial guess via forward simulation
x_init = zeros(nx, N);
x_init(:, 1) = x0;

for k=1:N-1
    x_init(:, k+1) = f(x_init(:, k));
end

for k=1:M-N
    x0 = x_predict(:, k);
    Q0 = P_predict(:, :, k);
    Y_current = Y(:, k:k+N);
    
    cost = @(X) objective(X, Y_current, f, h, Q, R, Q0, x0, true);
    
    x_opt = fmincon(cost, x_init);
    
    x_estimate_mhe(:, k) = x_opt(:, end);
    
    % initial guess for next MHE iteration
    x_init(:, 1:N-1) = x_opt(:, 2:N);
    x_init(:, N) = f(x_init(:, N-1));
end

%% Plot estimation results

figure(2); 

subplot(2,1,1); hold on; grid on;

plot(1:M, X(1,:), '-');
plot(1:M, x_estimate_kf(1,:), '.-');
plot(N:M-1, x_estimate_mhe(1,:), '.-');
legend('true', 'KF', 'MHE');

subplot(2,1,2); hold on; grid on;

plot(1:M, X(2,:), '-');
plot(1:M, x_estimate_kf(2,:), '.-');
plot(N:M-1, x_estimate_mhe(2,:), '.-');
legend('true', 'KF', 'MHE');

%save('data_lin.mat', 'X', 'Y');

